local cv_maprecords = CV_RegisterVar({
    name = "mpr_active",
    defaultvalue = "On",
    possiblevalue = CV_OnOff,
    flags = CV_NETVAR,
})

local cv_debuglog = CV_RegisterVar({
    name = "mpr_debug_log",
    defaultvalue = "Off",
    possiblevalue = CV_OnOff,
})

local function open_log()
    if not isserver then return end

    local file = io.open("mpr_log.txt", "r")
    
    if file then
        local otherfile = io.open("mpr_log.1.txt", "w")
        
        if otherfile then
            otherfile:write(file:read("*a"))
            otherfile:close()
        end
        
        file:close()
    end
    
    return io.open("mpr_log.txt", "w")
end

local log_file = open_log()
if log_file then log_file:setvbuf("line") end

local DEFAULT_PROFILE = "default"

-- Toggle so new record isn't added every frame
local race_ended = false

-- Table with all known records. Server-side only, isn't synched with clients
local times = {}

local hooks = {
    PreMapLoad = {}, -- Called before RecordLoad. Used by the io extension but can be used to manipulate records on mapload in general
    RecordLoad = {}, -- Like MapLoad hook but it also receives current record with map id
    NewRecord = {}, -- Called when player gets new record
}

-- Is mpr currently running?
local active = true

-- Current profile for records. Can be also something like "combi", "tech", etc
local profile = DEFAULT_PROFILE

-- Next profile to set
local nextprofile

-- Record for current map
local current_record
 
local nolog_warned = false
local function log(fmt, ...)
    -- Would be so fucking funny to error out because of log
    local ok, s = pcall(string.format, fmt, ...)
    
    if not ok then return end
    
    if isserver then
        if log_file then
            log_file:write(s, '\n')
        elseif not nolog_warned then
            print("\130WARNING:\128 Log file is not available for writing")
            nolog_warned = true
        end
    end

    if cv_debuglog.value then
        print("[MPR] "..s)
    end
end

local function addMPRHook(name, func)
    if hooks[name] == nil then error("invalid hook name '"..tostring(name).."'") end
    
    table.insert(hooks[name], func)
end

local function callMPRHooks(name, ...)
    assert(hooks[name]) -- Only called internally, so this is why assert
    
    for i = 1, #hooks[name] do
        hooks[name][i](...)
    end
end

local INVALID_PLAYER = {
    name = "???",
    skin = "sonic",
    skincolor = 1,
    restat = { speed = -1, weight = -1 },
}

local function playerEntry(p)
    local data = {}

    if not p.valid then
        return INVALID_PLAYER
    end
    
    data.name = p.name
    
    if p.mo and p.mo.valid then
        data.skin = p.mo.skin
    else
        data.skin = "sonic"
    end
    
    data.skincolor = p.skincolor
    
    data.restat = { speed = -1, weight = -1 }
    
	local skindata = skins[data.skin]
	
    if skindata and ((p.kartspeed ~= skindata.kartspeed) or (p.kartweight ~= skindata.kartweight)) then
        data.restat.speed = p.kartspeed
        data.restat.weight = p.kartweight
    end
    
    return data
end

-- Creates list of tables with player data needed for record
local function getPlayerList(players)
    local list = {}

    for i = 1, #players do
        table.insert(list, playerEntry(players[i]))
    end
    
    return list
end

-- Converts array of strings to a record.
local function arrayToRecord(arr, start)
    local record = {}
    
    local i = start or 1
    
    -- Convinience
    local function read()
        local str = arr[i]
        i = i + 1
        return str
    end
    
    local numplayers = tonumber(read())
    
    record.players = {}
    
    for j = 1, numplayers do
        local p = {}
        p.name = read()
        p.skin = read()
        p.skincolor = tonumber(read())
        p.restat = { speed = tonumber(read()), weight = tonumber(read()) }
        record.players[j] = p
    end
    
    record.time = tonumber(read())
    
    record.encore = tonumber(read())
    
    return record
end

local function recordToArray(record, noquote)
    local arr = {}
    
    -- Convinience
    local function write(val, quoted)
        local val = tostring(val)
        
        if quoted and not noquote then
            val = '"'..val:gsub('"', '')..'"'
        end
        
        table.insert(arr, val)
    end
    
    write(#record.players)
    
    for i = 1, #record.players do
        local p = record.players[i]
        write(p.name, true)
        write(p.skin, true)
        write(p.skincolor)
        write(p.restat.speed or -1)
        write(p.restat.weight or -1)
    end
    
    write(record.time)
    
    write(record.encore and 1 or 0)
    
    return arr
end

-- Make table with single record
local function buildRecord(time, players, encore)
    return {
        players = getPlayerList(players),
        time = time,
        encore = encore,
    }
end

local function extendRecord(record, players)
    for i = 1, #players do
        --local add = true
        
        if not players[i].valid then continue end
        
		-- Because tie your own records is funny
        --for j = 1, #record.players do
        --    if players[i].name == record.players[j].name then
        --        add = false
        --        break
        --    end
        --end
        
        --if add then
			table.insert(record.players, playerEntry(players[i]))
		--end
    end
    
    return record
end

local function buildEmptyRecord()
    return {
        players = {},
        time = INT32_MAX,
    }
end

local function isRecordValid(record)
    return record ~= nil and record.time ~= INT32_MAX and #record.players ~= 0
end

local function dumpRecord(record)
    local result = {"{ "}
    
    local function writefield(name, value)
        table.insert(result, name.."="..tostring(value or "nil")..", ")
    end
    
    writefield("time", record.time)
    writefield("encore", record.encore)
    
    if record.players then
        table.insert(result, "players={ ")
        
        for i = 1, #record.players do
            table.insert(result, "{ ")
            writefield("name", record.players[i].name)
            writefield("skin", record.players[i].skin)
            writefield("skincolor", record.players[i].skincolor)
            writefield("restat.speed", (record.players[i].restat or {}).speed)
            writefield("restat.weight", (record.players[i].restat or {}).weight)
            table.insert(result, " }, ")
        end
        
        table.insert(result, "}, ")
    else
        table.insert(result, "players=nil, ")
    end
    
    table.insert(result, "}")
    
    return table.concat(result, "")
end

-- Count players in game
local function statPlayers()
    local stat = {
        in_game = 0,
        racing = 0,
    }

    for player in players.iterate do
        if not player.spectator then
            stat.in_game = stat.in_game + 1
        end
        
        if not (player.exiting or (player.pflags & PF_TIMEOVER) or player.spectator or player.sinked) then
            stat.racing = stat.racing + 1
        end
    end

    return stat
end

-- Get best time and players that got it (multiple players in case of tie)
local function bestTime()
    local besttime
    local bestplayers = {}

    for player in players.iterate do
        if not ((player.pflags & PF_TIMEOVER) or player.spectator or player.sinked)
           and player.exiting and (besttime == nil or player.realtime <= besttime) then
            if besttime == player.realtime then
                table.insert(bestplayers, player)
            else
                bestplayers = {player}
            end

            besttime = player.realtime
        end
    end

    return besttime, bestplayers
end

-- Get record for specified map in specified profile. If in_profile is nil,
-- current profile is used
local function getRecord(mapid, in_profile)
    in_profile = in_profile or profile

    times[in_profile] = times[in_profile] or {}

    return times[in_profile][mapid]
end

-- Set record for map and specified/current profile
local function setRecord(mapid, record, in_profile)
    if record and #record.players == 0 then record = mpr.buildEmptyRecord() end

    in_profile = in_profile or profile

    times[in_profile] = times[in_profile] or {}

    times[in_profile][mapid] = record
end

-- If record we got is better than existing one, returns true and adds that
-- record to store. Otherwise, does nothing and returns false
local function tryAddRecord(mapid, time, players)
    local record = getRecord(mapid)

    if record == nil or time <= record.time then
        if record ~= nil and time == record.time then
            record = extendRecord(record, players)
        else
            record = buildRecord(time, players)
        end

        setRecord(mapid, record)
    end
end

-- Set current profile
local function setProfile(name)
    nextprofile = name
end

-- Get current profile
local function getProfile()
    return profile
end

-- Toggle if maprecords should be active (does nothing if mpr_active cvar is Off)
local function setActive(new_active)
    if cv_maprecords.value then active = new_active end
end

-- Is mpr active currently?
local function isActive()
    return cv_maprecords.value and active
end

local function getDefaultProfile()
    return DEFAULT_PROFILE
end

-- TODO - maybe expose times in a more safe way?
local function getTimesTable()
    return times
end

COM_AddCommand("__mpr_send_record", function(p, ...)
    if current_record ~= nil then
        CONS_Printf(p, ":3") -- :3
        return
    end
    
    current_record = arrayToRecord({...})
    callMPRHooks("RecordLoad", current_record)
end, COM_ADMIN)

local function sendRecord(record)
    -- Allow sending records
    current_record = nil
    
    if isserver then
        local arr = recordToArray(record)
        
        COM_BufInsertText(server, "__mpr_send_record "..table.concat(arr, ' '))
    end
end

addHook("ThinkFrame", function()
	if replayplayback then return end
    if not isActive() then return end
    
    local stat = statPlayers()
	
	-- Combi TA is a thing now
	local combi = hcombi and (hcombi.combi_on or hcombi.leveltime_on)
	
    -- Game went "time attack" mode with sneakers only, disable records
    if stat.in_game < (combi and 3 or 2) then
        mpr.log("Auto-disable mpr because only 1 player is left in game")
        active = false
        return
    end
    
    if G_BattleGametype() then
        mpr.log("Auto-disable mpr in battle mode")
        active = false
        return
    end

    if not race_ended then
        if stat.racing == 0 then
            race_ended = true
            
            local besttime, bestplayers = bestTime()
            
            local record = buildRecord(besttime, bestplayers, encoremode)

            if isserver then
                tryAddRecord(gamemap, besttime, bestplayers, encoremode)
            end
            
            if current_record == nil or record.time <= current_record.time then
                mpr.log("New record at map %s (#%d) in profile %s", G_BuildMapName(gamemap), gamemap, profile)
                callMPRHooks("NewRecord", record, bestplayers)
            end
        end
    end
end)

addHook("MapLoad", function(mapid)
    active = cv_maprecords.value

    race_ended = false

    if nextprofile ~= nil then
        profile = nextprofile
        nextprofile = nil
    end
    
    callMPRHooks("PreMapLoad", mapid)
    
    if isserver then
        current_record = getRecord(mapid) or buildEmptyRecord()
    end
    
    sendRecord(current_record)
end)

addHook("NetVars", function(net)
    active = net(active)
    race_ended = net(race_ended)
    current_record = net(current_record)
end)

-- Expose some of functions
rawset(_G, "mpr", {
    cv_maprecords = cv_maprecords,

    log = log,
    addHook = addMPRHook,
    arrayToRecord = arrayToRecord,
    recordToArray = recordToArray,
    dumpRecord = dumpRecord,
    sendRecord = sendRecord,
    buildRecord = buildRecord,
    buildEmptyRecord = buildEmptyRecord,
    isRecordValid = isRecordValid,
    getRecord = getRecord,
    setRecord = setRecord,
    setProfile = setProfile,
    getProfile = getProfile,
    setActive = setActive,
    isActive = isActive,
    getDefaultProfile = getDefaultProfile,
    getTimesTable = getTimesTable,
})